Canvas.isInSandboxMode = true;

let value = 0;
let dir = 1;
let colourScheme = 0;

let engine = {
  maxRPM: 7300,
  redlineRPM: 8000,
  idle: 800,
};
let turbo = 500;

const SHIFT_TIME = 0.15;
const MAX_GEAR = 4;
// Percentage of maxRPM (>1 means into redline)
const MAX_RPM_PERC = 1.05;

////////////////////////////////////////////////////////////////////////////////////////////////////

let tachoApp = new VelocityTachometerApp(150, document, null);
tachoApp.centre = new Vector(150, 300);
let boostApp = new VelocityBoostApp(75, document, null);
boostApp.centre = new Vector(225, 75);

let tachoApp2 = new FusionTachometerApp(150, document, null);
tachoApp2.centre = new Vector(475, 300);
let boostApp2 = new FusionBoostApp(75, document, null);
boostApp2.centre = new Vector(550, 75);

let tachoApp3 = new RingularityTachometerApp(150, document, null);
tachoApp3.centre = new Vector(800, 300)
let boostApp3 = new RingularityBoostApp(75, document, null);
boostApp3.centre = new Vector(875, 75);


let tachos = [tachoApp, tachoApp2, tachoApp3];
let fis = [boostApp, boostApp2, boostApp3];
let apps = [...tachos, ...fis];

tachos.forEach(tacho => tacho.sendInitialVehicleData(engine));
fis.forEach(fi => fi.sendInitialVehicleData(turbo));
apps.forEach(app => {
  app.recreate(false);
  app.drawBackgroundCanvas();
});

////////////////////////////////////////////////////////////////////////////////////////////////////

/**
 * Outline of what happens:
 *    T0:
 *      Blinking hazards
 *      Lights on
 *    T1:
 *      Hazards off
 *      Parking brake off
 *    T2:
 *      Engine is started -> RPM=[0->IDLE]
 *    T3:
 *      Engine Idles -> RPM~=IDLE
 *    T4:
 *      Throttle [0->1]
 *      Go through gears [1->MAX_GEAR]
 *      Increase speed
 *    T5:
 *      Brake [0->1]
 *      Throttle=0
 *      Go down gears [MAX_GEAR->1]
 *      Go down RPM [MAX_RPM->0]
 *    T6:
 *      Idle engine again -> RPM~=IDLE
 *      Release brake [1->0]
 *    T7:
 *      Engine turns off -> RPM=0
 */

const SECTIONS = {
  HAZARDS: 1,
  BREAK: 1.25,
  ENGINE_START: 2.25,
  ENGINE_STARTED: 2.75,
  BRAKING: 6.6,
  BREAK2: 9,
  ENGINE_OFF: 10,
  FINISH: 11,
};

function getDefaultData() {
  return {
    pBrakeEnabled: false,
    temp: 50,
    fuel: 0.75,
    currRPM: 0,
    speed: 0,
    gear: 0,
    throttle: 0,
    brake: 0,
    indicators: {
      left: 0,
      right: 0,
    },
    lightState: 0
  };
}

function getRPMNoise(x) {
  return Math.sin(x * 35) / 50;
}

let x = setInterval(() => {
  if (value > SECTIONS.FINISH) {
    clearInterval(x);
    return;
  }

  let { tachometer, boost } = handleSection(value);

  tachos.forEach(tacho => tacho.update(tachometer));
  fis.forEach(fi => fi.update(boost));
  apps.forEach(app => app.draw());

  // Slow down for "driving" section
  value += (value > SECTIONS.ENGINE_STARTED && value < SECTIONS.BRAKING ? 0.0035 : 0.005) * dir;
}, 10);

////////////////////////////////////////////////////////////////////////////////////////////////////
// Sections                                                                                       //
////////////////////////////////////////////////////////////////////////////////////////////////////

function handleSection(currValue) {
  if (currValue < SECTIONS.HAZARDS) {
    return blinkHazards(currValue);
  } else if (currValue < SECTIONS.BREAK) {
    return { tachometer: getDefaultData(), boost: { boost: 0 } };
  } else if (currValue < SECTIONS.ENGINE_START) {
    return startEngine(currValue - SECTIONS.BREAK);
  } else if (currValue < SECTIONS.ENGINE_STARTED) {
    return idleEngine(currValue - SECTIONS.ENGINE_START);
  } else if (currValue < SECTIONS.BRAKING) {
    return speedOff(currValue - SECTIONS.ENGINE_STARTED);
  } else if (currValue < SECTIONS.BREAK2) {
    return braking(currValue - SECTIONS.BRAKING);
  } else if (currValue < SECTIONS.ENGINE_OFF) {
    return idleEngineAfterBrake(currValue - SECTIONS.BREAK2);
  } else {
    return turnOffEngine(currValue - SECTIONS.ENGINE_OFF);
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////
// Section: Blink Hazards                                                                         //
////////////////////////////////////////////////////////////////////////////////////////////////////

function blinkHazards(currValue) {
  let indicatorState =  Math.floor((3 * currValue) + 1) % 2;
  return {
    tachometer: {
      ...getDefaultData(),
      pBrakeEnabled: true,
      indicators: {
        left: indicatorState,
        right: indicatorState,
      },
      lightState: 1,
    },
    boost: {
      boost: 0
    }
  };
}

////////////////////////////////////////////////////////////////////////////////////////////////////
// Section: Start Engine                                                                          //
////////////////////////////////////////////////////////////////////////////////////////////////////

function startEngine(x) {
  let rpm = x < 0.1 ? 10 * x : 1 + getRPMNoise(x);
  return {
    tachometer: {
      ...getDefaultData(),
      currRPM: rpm * engine.idle
    },
    boost: {
      boost: 0
    }
  };
}

////////////////////////////////////////////////////////////////////////////////////////////////////
// Section: Idle Engine                                                                           //
////////////////////////////////////////////////////////////////////////////////////////////////////

function idleEngine(x) {
  return {
    tachometer: {
      ...getDefaultData(),
      currRPM: engine.idle * (1 + getRPMNoise(x))
    },
    boost: {
      boost: 0
    }
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////
// Section: Speed Off                                                                             //
////////////////////////////////////////////////////////////////////////////////////////////////////

/**
 * https://www.desmos.com/calculator/pnkbxvczoa
 * Given:
 *  [0 <= x <= MAX_GEAR],
 *  [0 < g <= MAX_GEAR], g € Integer
 */
function getRPM(x, g) {
  const IDLE_PERC = engine.idle / engine.maxRPM;
  let s = ( (g - 1) / (1.5 * MAX_GEAR) ) + IDLE_PERC;
  let e = (  g      / (1.5 * MAX_GEAR) ) + IDLE_PERC;

  if (x < g - SHIFT_TIME) return ( ( (MAX_RPM_PERC - s) * (x + 1 - g) ) / (1 - SHIFT_TIME) ) + s;
  return ( ( (e - MAX_RPM_PERC) * (x - g) ) / SHIFT_TIME ) + e;
}

function getSpeed(x, g) {
  let speed = (x) => 1 - Math.exp(-x);
  const GEAR1_END = speed(1 - SHIFT_TIME);

  if (x < g - SHIFT_TIME) return speed(x - g + 1) + (g - 1) * GEAR1_END;
  return g * GEAR1_END;
}

function getBoost(x, g) {
  if (x < g - SHIFT_TIME) return Math.min((x - g + 1) * 5, 1) * turbo;
  return Math.max((g - x) * 10, 0) * turbo;
}

function speedOff(x) {
  // The "part-way through" gear
  // (gear=1, decimalGear=1.5) means halfway through gear 1
  let decimalGear = x + 1;
  let gear = Math.floor(decimalGear);

  return {
    tachometer: {
      ...getDefaultData(),
      currRPM: getRPM(x, gear) * engine.maxRPM,
      speed: getSpeed(x, gear) * 30,
      // Gear up occurs at prevGear.(1 - SHIFT_TIME) [such as 2.9], but currRPM occurs on the integer [such as 3]
      gear: decimalGear - gear >= 1 - SHIFT_TIME ? gear + 1 : gear,
      throttle: Math.min(x * 5, 1)
    },
    boost: {
      boost: getBoost(x, gear)
    }
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////
// Section: Brake                                                                                 //
////////////////////////////////////////////////////////////////////////////////////////////////////

function braking(x) {
  let sectionLenth = SECTIONS.BREAK2 - SECTIONS.BRAKING;
  let percX = x / sectionLenth;
  let invPercX = 1 - percX;

  let gear = Math.floor((MAX_GEAR * invPercX)) + 1;

  return {
    tachometer: {
      ...getDefaultData(),
      currRPM: getRPM(invPercX * 3.85, gear) * engine.maxRPM,
      speed: Math.pow((percX - 1), 2) * 68,
      gear: gear,
      brake: Math.min(x * 3, 1)
    },
    boost: {
      boost: Math.max((0.2 - x) * 5 * turbo, 0)
    }
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////
// Section: Idle Engine After Brake                                                               //
////////////////////////////////////////////////////////////////////////////////////////////////////

function idleEngineAfterBrake(x) {
  let vals = idleEngine(x);
  vals.tachometer.brake = Math.max(1 - (x * 5), 0);
  return vals;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
// Section: Turn Off Engine                                                                       //
////////////////////////////////////////////////////////////////////////////////////////////////////

function turnOffEngine(x) {
  return {
    tachometer: {
      ...getDefaultData(),
      currRPM: Math.max(0, 1 - (x * 7.5)) * engine.idle,
    },
    boost: { boost: 0 }
  }
}