angular.module("beamng.apps").directive("fusionTacho", [
  function () {
    return {
      template:
        '<div id="FT_container">' +
        '<link rel="stylesheet" href="../../../resources/base.css">' +
        '<link rel="stylesheet" href="../../modules/apps/FusionTacho/app.css">' +
        '<canvas class="canvas" id="FT_bgCanvas" width="300" height="300"></canvas>' +
        '<canvas class="canvas" id="FT_canvas"   width="300" height="300"></canvas>' +
        "</div>",
      replace: true,
      restrict: "EA",
      link: function ($scope, $element) {
        // This is all that is variable between apps
        const APP_TYPE = "tachometer";
        const APP_THEME = "fusion";

        //////////////////////////////////////////////////////////////////////////////////////////////
        // Everything Below Required In All Apps                                                    //
        // Cannot be extracted into another file as cannot be loaded by Angular before then         //
        //////////////////////////////////////////////////////////////////////////////////////////////

        let vars = {
          /**
           * @type AbstractAppContainer
           */
          appContainer: null,
          /**
           * @type string[]
           */
          requiredStreams: null,
          /**
           * @type NullableDimensions
           */
          initialDimensions: null
        };

        ////////////////////////////////
        // Load the base file //
        ////////////////////////////////

        /**
         * @returns {boolean} Whether the main script has loaded yet.
         */
        function hasMainScriptLoaded() {
          return typeof AbstractAppContainer !== "undefined";
        }

        /**
         * Waits for the main script and then runs the given function when it has loaded.
         * If the main script does not load after 5s, an error is logged and the function remains uncalled.
         * @param {Function} func - the function to run.
         */
        function waitForMainScript(func) {
          // Poll every 1s
          let t = setInterval(() => {
            if (hasMainScriptLoaded()) {
              func();
              clearInterval(t);
              t = null;
            }
          }, 1000);

          // Give up after 5s; assume it is not going to load after then
          setTimeout(() => {
            if (t) {
              console.warn("Attempted to load class AbstractAppContainer; timed out! The app may not need to be loaded...");
              clearInterval(t);
            }
          }, 5000);
        }

        /**
         * Creates a new app instance.
         * @param {class<T>} appClass       - the app class to create an instance of.
         * @param {Object}   bngApi         - the BeamNG api service.
         * @param {Object}   UiUnits        - the BeamNG UiUnits service.
         * @param {number}   radius         - the default radius of the app.
         * @param {class<S>} containerClass - the container class to wrap the app in.
         * @returns {S}
         * @template T: ? extends {@link AbstractBaseApp}
         * @template S: ? extends {@link AbstractAppContainer}
         */
        function loadApp(appClass, bngApi, UiUnits, radius, containerClass) {
          let app = new appClass(radius, document, bngApi, UiUnits);
          return new containerClass(app);
        }

        /**
         * Creates a new tachometer app instance, wrapped in a container class.
         * @param {class<T>} appClass - the app class to create an instance of.
         * @param {Object}   bngApi   - the BeamNG api service.
         * @param {Object}   UiUnits  - the BeamNG UiUnits service.
         * @returns {TachometerAppContainer} The. App container created.
         */
        function loadTachometerApp(appClass, bngApi, UiUnits) {
          return loadApp(appClass, bngApi, UiUnits, 150, TachometerAppContainer);
        }

        /**
         * Creates a new forced induction app instance, wrapped in a container class.
         * @param {class<T>} appClass - the app class to create an instance of.
         * @param {Object}   bngApi   - the BeamNG api service.
         * @param {Object}   UiUnits  - the BeamNG UiUnits service.
         * @returns {BoostAppContainer} The. App container created.
         */
        function loadBoostApp(appClass, bngApi, UiUnits) {
          return loadApp(appClass, bngApi, UiUnits, 75, BoostAppContainer);
        }

        /**
         * Returns the app class for an app type and theme.
         * Note that the scripts MUST be loaded in order for the returned object to exist.
         * @param {string} appType  - the app type.
         * @param {string} appTheme - the app theme.
         * @returns {AbstractBaseApp} The. App class.
         */
        function getAppClass(appType, appTheme) {
          switch (true) {
            case appTheme === "velocity" && appType === "tachometer":
              return VelocityTachometerApp;
            case appTheme === "velocity" && appType === "forcedInduction":
              return VelocityBoostApp;
            case appTheme === "fusion" && appType === "tachometer":
              return FusionTachometerApp;
            case appTheme === "fusion" && appType === "forcedInduction":
              return FusionBoostApp;
            case appTheme === "ringularity" && appType === "tachometer":
              return RingularityTachometerApp;
            case appTheme === "ringularity" && appType === "forcedInduction":
              return RingularityBoostApp;
          }
        }

        /**
         * Loads the additional scripts required for this app to function.
         * @param {string}                                               appType  - the app type.
         * @param {string}                                               appTheme - the app theme.
         * @param {Object}                                               bngApi   - the BeamNG api service.
         * @param {Object}                                               UiUnits  - the BeamNG UiUnits service.
         * @param {(AbstractAppContainer, AbstractAppContainer) => void} callback - the callback function to call once the
         *                                                                        app is setup.
         */
        function createApp(appType, appTheme, bngApi, UiUnits, callback) {
          let isTachometerApp = appType === "tachometer";
          let appClass = getAppClass(appType, appTheme);

          let app = isTachometerApp
            ? loadTachometerApp(appClass, bngApi, UiUnits)
            : loadBoostApp(appClass, bngApi, UiUnits);
          let containerClass = isTachometerApp ? TachometerAppContainer : BoostAppContainer;

          callback(app, containerClass);
        }

        /**
         * Adds in the tachometersAppHandler script for loading everything else.
         * Acts as an interface between the app directive and the app logic.
         * @param {string} appType
         * - the app type.
         * @param {string} appTheme
         * - the app theme.
         * @param {Object} bngApi
         * - the BeamNG api service.
         * @param {Object} UiUnits
         * - the BeamNG UiUnits service.
         * @param {(appCon: AbstractAppContainer, containerClass: class extends AbstractAppContainer) => void} callback
         * - the callback function to call once the app is setup.
         */
        function loadMainScript(appType, appTheme, bngApi, UiUnits, callback) {
          const ID = "tachometers";
          let task = () => createApp(appType, appTheme, bngApi, UiUnits, callback);

          let script = document.getElementById(ID);
          // If the script already exists, don't add it in again
          if (script) {
            if (!hasMainScriptLoaded()) {
              waitForMainScript(task);
            } else {
              task();
            }
          } else {
            script = document.createElement("script");
            script.src = "../../composite.all.js";
            script.id = ID;
            script.onload = task;
            document.head.append(script);
          }
        }

        // noinspection JSClosureCompilerSyntax
        /**
         * Called when the app loads.
         * @param {AbstractAppContainer}        appCon         - the app container instance.
         * @param {class<AbstractAppContainer>} containerClass - the app container's class.
         */
        function onAppLoaded(appCon, containerClass) {
          vars.appContainer = appCon;
          vars.requiredStreams = appCon.getRequiredStreams();
          StreamsManager.add(vars.requiredStreams);
        }

        /**
         * Checks whether initial dimensions have been stored and whether the app is available. If so,
         * this function sets those initial dimensions and then clears them. Else, this function does
         * nothing
         */
        function checkAndSetInitDims() {
          if (vars.initialDimensions && vars.appContainer) {
            // Apply initial dimensions
            vars.appContainer.onAppResized(vars.initialDimensions);
            vars.initialDimensions = null;
          }
        }

        // INIT
        loadMainScript(APP_TYPE, APP_THEME, bngApi, UiUnits, onAppLoaded);

        ////////////
        // Events //
        ////////////

        $scope.$on("VehicleFocusChanged", () => {
          if (vars.appContainer) {
            vars.appContainer.hideApp();
            vars.initialDimensions = vars.appContainer.getCurrentAppDimensions();
          }

          loadMainScript(APP_TYPE, APP_THEME, bngApi, UiUnits, onAppLoaded);
        });

        $scope.$on("$destroy", () => {
          if (vars.requiredStreams) StreamsManager.remove(vars.requiredStreams);
        });

        $scope.$on("streamsUpdate", (_event, streams) => {
          checkAndSetInitDims();

          if (vars.appContainer) vars.appContainer.onStreamsUpdated(streams);
        });

        $scope.$on("app:resized", (_event, dimensions) => {
          if (vars.appContainer) vars.appContainer.onAppResized(dimensions);
          // Important not to overwrite any existing initial dimensions, since switching vehicle takes
          // precedence over this, so if they have been set by that function, don't overwrite them
          else if (!vars.initialDimensions)
            // Save for when the app is loaded
            vars.initialDimensions = dimensions;
        });

        //////////////////////////////////////////////////////////////////////////////////////////////
      }
    };
  }
]);
