/**
 * A class representing a 2D vector
 */
export class Vector {
  private readonly array: number[];

  /**
   * @class
   * @param {number} [x] - value of x dimension. Defaults to 0.
   * @param {number} [y] - value of y dimension. Defaults to 0.
   */
  constructor(x = 0, y = 0) {
    this.array = [x, y];
  }

  /**
   * Copies a vector and returns its copy.
   * @param {Vector} v - the vector to copy.
   * @returns {Vector} The copied vector instance.
   */
  public static copy(v: Vector): Vector {
    return new Vector(v.array[0], v.array[1]);
  }

  /**
   * Rotates this vector around the origin by a given angle amount.
   * This does not affect this vector and a new vector is returned.
   * @param {number} angle - the angle (in radians)
   * @returns {Vector} The resultant vector.
   */
  public rotateAboutOrigin(angle: number): Vector {
    let cos = Math.cos(angle);
    let sin = Math.sin(angle);
    return new Vector(this.x() * cos - this.y() * sin, this.x() * sin + this.y() * cos);
  }

  /**
   * @param {Vector}                           vector
   * @param {Vector | number}                  amount
   * @param {(a: number, b: number) => number} func
   * @returns {Vector}
   */
  private static mapOperations(
    vector: Vector,
    amount: Vector | number,
    func: (a: number, b: number) => number
  ): Vector {
    let result =
      amount instanceof Vector
        ? vector.array.map((v, i) => func(v, amount.array[i]))
        : vector.array.map(e => func(e, amount));

    return new Vector(result[0], result[1]);
  }

  /**
   * Clones this vector and returns its copy.
   * @returns {Vector} The cloned vector.
   */
  public clone(): Vector {
    return Vector.copy(this);
  }

  /**
   * Clones this vector, and adds to another vector or scalar value and returns the result.
   * @param {Vector | number} amount - the vector/scalar to add.
   * @returns {Vector}
   */
  public add(amount: Vector | number): Vector {
    return Vector.mapOperations(this, amount, (e, a) => e + a);
  }

  /**
   * Clones this vector, and subtracts to another vector or scalar value and returns the result.
   * @param {Vector | number} amount - the vector/scalar to subtract.
   * @returns {Vector}
   */
  public sub(amount: Vector | number): Vector {
    return Vector.mapOperations(this, amount, (e, a) => e - a);
  }

  /**
   * Clones this vector, and multiplies to another vector or scalar value and returns the result.
   * @param {Vector | number} amount - the vector/scalar to subtract.
   * @returns {Vector}
   */
  public mult(amount: Vector | number): Vector {
    return Vector.mapOperations(this, amount, (a, b) => a * b);
  }

  /**
   * Clones this vector, and divides to another vector or scalar value and returns the result.
   * @param {Vector | number} amount - the vector/scalar to divide with.
   * @returns {Vector}
   */
  public div(amount: Vector | number): Vector {
    return Vector.mapOperations(this, amount, (e, a) => e / a);
  }

  public addX(incX: number): Vector {
    return new Vector(this.x() + incX, this.y());
  }

  public addY(incY: number): Vector {
    return new Vector(this.x(), this.y() + incY);
  }

  public x(): number {
    return this.array[0];
  }

  public y(): number {
    return this.array[1];
  }

  public toString(): string {
    return `Vector[x=${this.x()}, y=${this.y()}]`;
  }
}
